/*
   Utility to decrypt Shiva crypt blocks
   Copyright (c) 2003, Chris Eagle
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option) 
   any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
   more details.
   
   You should have received a copy of the GNU General Public License along with 
   this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
   Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <idc.idc>

/*
 * You need to enter all of the Shiva key data here
 * Its ugly, but it works.
 *
 * Run the script by placing the IDA cursor on the
 * Magic value for the first crypt block, then invoking
 * the script.
 */

#define TYPE_0_0 0xF77BFD6F
#define TYPE_0_1 0x69BEFF33
#define TYPE_0_2 0xFF5FDF87
#define TYPE_0_3 0x6D1CB4BC

#define TYPE_1_0 0xDA0E6044
#define TYPE_1_1 0x2BE08AEE
#define TYPE_1_2 0xD524E991
#define TYPE_1_3 0xC54BF4A4

#define TYPE_2_0 0xB6772009
#define TYPE_2_1 0x56BB54E3
#define TYPE_2_2 0x71AA986F
#define TYPE_2_3 0x03A345FC

#define TYPE_3_0 0x03A345FC
#define TYPE_3_1 0xA0D657A9
#define TYPE_3_2 0x6307DCE9
#define TYPE_3_3 0x5A4A9D9A

#define TYPE_4_0 0x71EEFDDC
#define TYPE_4_1 0xB4F733A1
#define TYPE_4_2 0xD85A325F
#define TYPE_4_3 0x7E5FED28

static TEA_encrypt(source, dest, key) {
   auto n, delta, sum, y, z;
   auto p1, p2, p3, p4, p5;
   n = 32;
   delta = 0x9E3779B9;
   sum = 0;
   y = GetArrayElement(AR_LONG, source, 0);
   z = GetArrayElement(AR_LONG, source, 1);
   while (1) {
      if (n <= 0) break;
      n--;
      y = y + ((z << 4 ^ ((z >> 5) & 0x07FFFFFF)) + z ^ sum 
               + GetArrayElement(AR_LONG, key, sum & 3));
      sum = sum + delta;
      z = z + ((y << 4 ^ ((y >> 5) & 0x07FFFFFF)) + y ^ sum 
               + GetArrayElement(AR_LONG, key, (sum >> 11) & 3));
   }
   SetArrayLong(dest, 0,  y);
   SetArrayLong(dest, 1,  z);
}

static TEA_decrypt(source, dest, key) {
   auto n, y, z, sum, delta;
   n = 32;
   delta = 0x9E3779B9;
   sum = 0xC6EF3720;
   y = GetArrayElement(AR_LONG, source, 0);
   z = GetArrayElement(AR_LONG, source, 1);
   while (1) {
      if (n <= 0) break;
      n--;
      z = z - ((y << 4 ^ ((y >> 5) & 0x07FFFFFF)) + y ^ sum 
               + GetArrayElement(AR_LONG, key, (sum >> 11) & 3));
      sum  = sum - delta;
      y = y - ((z << 4 ^ ((z >> 5) & 0x07FFFFFF)) + z ^ sum 
                + GetArrayElement(AR_LONG, key, sum & 3));
   }
   SetArrayLong(dest, 0,  y);
   SetArrayLong(dest, 1,  z);
}

static buildKeys() {
   auto keys, row;
   
   DeleteArray(GetArrayId("keys"));
   keys = CreateArray("keys");
   
   DeleteArray(GetArrayId("row0"));
   row = CreateArray("row0");
   SetArrayLong(row, 0, TYPE_0_0);
   SetArrayLong(row, 1, TYPE_0_1);
   SetArrayLong(row, 2, TYPE_0_2);
   SetArrayLong(row, 3, TYPE_0_3);
   SetArrayLong(keys, 0, row);

   DeleteArray(GetArrayId("row1"));
   row = CreateArray("row1");
   SetArrayLong(row, 0, TYPE_1_0);
   SetArrayLong(row, 1, TYPE_1_1);
   SetArrayLong(row, 2, TYPE_1_2);
   SetArrayLong(row, 3, TYPE_1_3);
   SetArrayLong(keys, 1, row);

   DeleteArray(GetArrayId("row2"));
   row = CreateArray("row2");
   SetArrayLong(row, 0, TYPE_2_0);
   SetArrayLong(row, 1, TYPE_2_1);
   SetArrayLong(row, 2, TYPE_2_2);
   SetArrayLong(row, 3, TYPE_2_3);
   SetArrayLong(keys, 2, row);

   DeleteArray(GetArrayId("row3"));
   row = CreateArray("row3");
   SetArrayLong(row, 0, TYPE_3_0);
   SetArrayLong(row, 1, TYPE_3_1);
   SetArrayLong(row, 2, TYPE_3_2);
   SetArrayLong(row, 3, TYPE_3_3);
   SetArrayLong(keys, 3, row);

   DeleteArray(GetArrayId("row4"));
   row = CreateArray("row4");
   SetArrayLong(row, 0, TYPE_4_0);
   SetArrayLong(row, 1, TYPE_4_1);
   SetArrayLong(row, 2, TYPE_4_2);
   SetArrayLong(row, 3, TYPE_4_3);
   SetArrayLong(keys, 4, row);

   return keys;
}

static freeKeys(keys) {
   DeleteArray(GetArrayElement(AR_LONG, keys, 0));
   DeleteArray(GetArrayElement(AR_LONG, keys, 1));
   DeleteArray(GetArrayElement(AR_LONG, keys, 2));
   DeleteArray(GetArrayElement(AR_LONG, keys, 3));
   DeleteArray(GetArrayElement(AR_LONG, keys, 4));
   DeleteArray(keys);
}

static TEAblock(addr, len, keyrow) {
   auto source, dest, i, keybyte;

   DeleteArray(GetArrayId("source"));
   source = CreateArray("source");
   DeleteArray(GetArrayId("dest"));
   dest = CreateArray("dest");

//   for (i = 0; i < 4; i = i + 1) {
//      Message("Key[%d] = 0x%X\n", i, GetArrayElement(AR_LONG, keyrow, i));
//   }
   
   MakeUnkn(addr, 1);
   
   while (len > 7) {
      SetArrayLong(source, 0, Dword(addr));
      SetArrayLong(source, 1, Dword(addr + 4));
      TEA_decrypt(source, dest, keyrow);
      PatchDword(addr, GetArrayElement(AR_LONG, dest, 0));
      PatchDword(addr + 4, GetArrayElement(AR_LONG, dest, 1));
      addr = addr + 8;
      len = len - 8;
   }
   if (len != 0) {
      SetArrayLong(source, 0, 0x61616161);
      SetArrayLong(source, 1, 0x61616161);
      TEA_encrypt(source, dest, keyrow);
      //dest now holds bytes for decrypting excess
      keybyte = GetArrayElement(AR_LONG, dest, 0);
      i = 0;
      while (i < len) {
         PatchByte(addr, Byte(addr) ^ (keybyte & 0xFF));
         addr = addr + 1;
         keybyte = keybyte >> 8;
         i = i + 1;         
         if (i == 4) {
            keybyte = GetArrayElement(AR_LONG, dest, 1);
         }
      }
   }
   
   DeleteArray(source);
   DeleteArray(dest);
}
         
static main(void) {
   auto start, ecx, nBytes, ebx, size;
   auto edx, edi, msg, addr, nItems;
   auto keys, magic;

   keys = buildKeys();

   start = here;
   addr = start;

   while (1) {
      magic = Dword(addr);  //magic
      magic = (magic >> 16) & 0xF;
      addr = addr + 4;

      size = Dword(addr) - 8;
      addr = addr + 4;
      TEAblock(addr, size, GetArrayElement(AR_LONG, keys, magic));

      nItems = size / 4;
      MakeDword(addr);  //first item
      MakeArray(addr, nItems);

      addr = addr + size;

      if (magic == 4) break; //all done
   }

   freeKeys(keys);
   
   Message("Crypt blocks range [0x%X - 0x%X)\n", start, addr);
   
}
