/*
 * XTEA (Tiny Encryption Algorithm)
 *
 * This is one of the most simple and fastest encryption methods, and it is in
 * theory as secure as IDEA and other complex ciphers. 
 *
 * TEA was developed by David Wheeler and Roger Needham, but both the algorithm
 * and this implementation in ANSI C where released to Public Domain.
 */

typedef unsigned int uint;
typedef unsigned char uchar;

void TEA_encrypt(uint source[], uint dest[], uint key[]) {
   int n = 32;
   uint delta = 0x9E3779B9;
   uint sum = 0;
   uint y = source[0];
   uint z = source[1];
   while (n-- > 0) {
      y += (z << 4 ^ z >> 5) + z ^ sum + key[sum & 3];
      sum += delta;
      z += (y<<4 ^ y>>5) + y ^ sum + key[sum>>11 &3];
   }
   dest[0] = y;
   dest[1] = z;
}

void TEA_decrypt(uint source[], uint dest[], uint key[]) {
   int n = 32;
   uint delta = 0x9E3779B9;
   uint sum = 0xC6EF3720;
   uint y = source[0];
   uint z = source[1];
   while (n-- > 0) {
      z -= (y << 4 ^ y >> 5) + y ^ sum + key[sum >> 11 & 3];
      sum -= delta;
      y -= (z << 4 ^ z >> 5) + z ^ sum + key[sum & 3];
   }
   dest[0] = y;
   dest[1] = z;
}

void TEA_block_decrypt(uchar *addr, uint len, uint key[]) {
   uint *data = (uint*) addr;
   while (len > 7) {
      TEA_decrypt(data, data, key);
      data += 2;
      len = len - 8;
   }
   if (len) {
      uint xKey[] = {0x61616161, 0x61616161};
      uchar *keyByte = (uchar*) xKey;
      TEA_encrypt(xKey, xKey, key);
      //dest now holds bytes for decrypting excess
      addr = (uchar*) data;
      for (int i = 0; i < len; i++) {
         *addr = *addr ^ *keyByte++;
         addr++;
      }
   }
}

